// src/backend/text_simulation.hh - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#ifndef TEXT_SIMULATION_HH
#define TEXT_SIMULATION_HH 1

#include "config.h"
#include "gettext.h"

#include "backend/simulation.hh"
#include "io_manager.hh"
#include "templates/smartp.hh"
//#include "backend/policy_parameters.hh"
#include "backend/string_utils.hh"

#include "smartp.hh"

#include <glibmm/thread.h>
#include <glibmm/ustring.h>
#include <utility>
#include <map>

namespace sgpem
{
  template <typename T>
  class CommandParameter;
}

namespace sgpem
{
  class TextSimulation;

  /**
    \brief Concrete \ref Simulation subclass with a text-based user interface.

    A command-based interface is used, so methods of the base class can be called with
    a proper command string obtained from the input device(s).
    Any object returned after the call to Simulation will be returned to the output
    devices(s) in a human-readable format.
   */
  class SG_DLLEXPORT TextSimulation : public sigc::trackable
  {
  public:
    ~TextSimulation();

    /**
      \brief Executes a command read from an input device.

      A list of supported commands follows:
      
      \li help \<string\> <br>
      If \<string\> is a valid command, usage instructions are printed 
      for the command.

      \li run <br>
      Calls run()

      \li pause <br>
      Calls pause()

      \li stop <br>
      Calls stop()

      \li setmode \<bool\> <br>
      Calls set_mode()

      \li getmode <br>
      Calls get_mode()

      \li settimer \<int\> <br>
      Calls set_timer()

      \li gettimer <br>
      Calls get_timer()
      
      \li reset <br>
      Calls reset()

      \li jumpto \<int\> <br>
      Calls jump_to()

      \li setpolicy \<string\> <br>
      Calls set_policy()

      \li getpolicy <br>
      Calls get_policy(), and prints the name and a description of 
      the returned policy

      \li getpolicies <br>
      Calls get_avaiable_policies(), and prints a list of policy 
      descriptions in the same way as getpolicy does

      \li setpolicyattributes \<name\> = \<value\>; \<name\> = \<value\>; <br>
      Changes the value of the policy's attributes
      
      \li  getpolicyattributes <br>
      Prints the name and the value of the policy's attributes
     */
    static void parse_command(TextSimulation &sim, const Glib::ustring& str);

    /**
      Adds an available I/O device.
     */
    void add_io_device(memory::smart_ptr<IOManager>);

    /**
      Prints the actual state of the simulation, with emphasis on the current
      status of the scheduling process (ready queue and running process).
     */
    void update();

  private:
    bool check_arguments_num(const Tokens& arguments, unsigned int num);
    template <typename Container>
    void show(const Container& entities);
    template <typename T>
    void get_parameter(CommandParameter<T>& parameter);
    void on_run(const Tokens& arguments);
    void on_pause(const Tokens& arguments);
    void on_stop(const Tokens& arguments);
    void on_configure_cpu_policy(const Tokens& arguments);
    void on_help(const Tokens& arguments);
    void on_quit(const Tokens& arguments);  
    void on_get(const Tokens& arguments);
    void on_set(const Tokens& arguments);
    void on_show(const Tokens& arguments);
    void on_show_processes(const Tokens& arguments);
    void on_show_resources(const Tokens& arguments);
    void on_show_threads(const Tokens& arguments);
    void on_show_requests(const Tokens& arguments);
    void on_show_subrequests(const Tokens& arguments);
    void on_show_cpu_policies(const Tokens& arguments);
    void on_show_resource_policies(const Tokens& arguments);
    void on_add(const Tokens& arguments);
    void on_add_process(const Tokens& arguments);
    void on_add_resource(const Tokens& arguments);
    void on_add_thread(const Tokens& arguments);
    void on_add_request(const Tokens& arguments);
    void on_add_subrequest(const Tokens& arguments);
    void on_remove(const Tokens& arguments);  
    void on_remove_process(const Tokens& arguments);
    void on_remove_resource(const Tokens& arguments);
    void on_remove_thread(const Tokens& arguments);
    void on_remove_request(const Tokens& arguments);
    void on_remove_subrequest(const Tokens& arguments);
    
    // FIXME This is a temporary replacement for the 
    // to-be written I/O layer
    static void p_stdout(const Glib::ustring& str);
    static void p_stderr(const Glib::ustring& str);
    static Glib::ustring readline();
    
    std::vector<memory::smart_ptr<IOManager> > _devices;
    static void _io_loop(std::pair<TextSimulation*, int>);
  };

}


#endif
